#!/usr/bin/env python
# -*- coding: utf-8; -*-
#
#
# Description
# Finite State Machine Class
# Class with the following functions:
# 1. turn on/off states from spike sources OK
# 2. turn on states from states with half connections OK
# 3. turn on/off states from states with full connections OK
#

import os
import sys
import numpy as np
	
from pyhalbe import HICANN
import pyhalbe.Coordinate as C
from pysthal.command_line_util import init_logger
import pysthal

import pyhmf as pynn
from pymarocco import PyMarocco, Defects
from pymarocco.runtime import Runtime
from pymarocco.coordinates import LogicalNeuron
from pymarocco.results import Marocco
import pylogging

class FSA_BrainScaleS: 
    def __init__(self):
       self.CA_SIZE = 20
       self.CA_INHIBS = 5
       self.numberHalfSynapses = 6
       self.numSpikeSources = 15
       self.fullWeight = 6
       self.intraWeight = 14
       self.inputWeight = 15
       self.inhibWeight = 11
       self.sourceInhibWeight = 15
       self.halfWeight = 2

       self.neuron_parameters = {'v_thresh':-10,'v_reset':-40.0, 
            'tau_refrac': 1.0,'tau_syn_E': 20,'tau_syn_I': 5,
            'v_rest':-20,'e_rev_I': -50., 'cm': 1.0,
            'e_rev_E':8.,'tau_m': 10,'i_offset':0}       

    # allocate a population of neurons for the states
    def allocateNeurons(self, numNeurons):
        pop = pynn.Population(numNeurons, pynn.IF_cond_exp, 
                              self.neuron_parameters)
        return pop

    #run the simulator
    def runSim(self, duration):
        pynn.run(duration)

    #reset and the end of simulation to support downloading spiking and voltage.
    def resetSim(self):
        pynn.reset()

    def printParameters(self, population):
        vt = population.get('v_thresh')
        vre = population.get('v_reset')
        tr = population.get('tau_refrac')
        tsE = population.get('tau_syn_E')
        tsI = population.get('tau_syn_I')
        vrest = population.get('v_rest')
        cm = population.get('cm')
        erE = population.get('e_rev_E')
        erI = population.get('e_rev_I')
        tm = population.get('tau_m')

        print 'vthresh', vt[0]
        print 'v_reset', vre[0]
        print 'tau_refrac', tr[0]
        print 'tau_syn_E', tsE[0]
        print 'tau_syn_I', tsI[0]
        print 'v_rest', vrest[0]
        print 'cm', cm[0]
        print 'e_rev_E', erE[0]
        print 'e_rev_I', erI[0]
        print 'tau_m', tm[0]


    #Call pynn to get a spike generator.  Pass in an array of times to spike.
    def getSpikeSources(self,spikeTimes):
        spikeSourceArray = pynn.Population(self.numSpikeSources, pynn.SpikeSourceArray,{ 'spike_times': spikeTimes})
        return spikeSourceArray

    ##NB all pynn connections have to be reset in hardware.  This class
    ##uses setBrainScaleSWt for that.
    def turnOnStateFromSpikeSource(self,spikeSource, toNeurons, toCA):
        #print "input weight = ", self.inputWeight
        connector = []
        for fromOffset in range (0,self.numSpikeSources):
            for toOffset in range (0, self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromOffset,toNeuron, self.inputWeight,1.0)]
        inputConnList = pynn.FromListConnector(connector)
        projections = [
        pynn.Projection(spikeSource, toNeurons, inputConnList, target='excitatory'),
        ]
        return projections[0]

    def turnOffStateFromSpikeSource(self,spikeSource, toNeurons, toCA):
        #print "input weight = ", self.inputWeight
        #brainScales constant
        connector = []
        for fromOffset in range (0,self.numSpikeSources):
            for toOffset in range (0, self.CA_SIZE-self.CA_INHIBS):
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromOffset,toNeuron,
                                          self.sourceInhibWeight,1.0)]
        inputConnList = pynn.FromListConnector(connector)
        projections = [
            pynn.Projection(spikeSource, toNeurons, inputConnList, target='inhibitory'),
        ]

        return projections[0]

    def stateStimulatesState(self, fromNeurons, fromCA, toNeurons,toCA,numberConnections,weight):
        connector = []
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            for toSynapse in range (0,numberConnections):
                fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
                toOffset = toSynapse + fromOffset + 1
                toOffset = toOffset%(self.CA_SIZE-self.CA_INHIBS)
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,weight,1.0)]
        halfConnList = pynn.FromListConnector(connector)
        projections = [
            pynn.Projection(fromNeurons, toNeurons, halfConnList, target='excitatory'),
        ]

        return projections[0]

    ##Setup connections from one state to another so that one firing pre-CA
    ##is insufficient to start the post, but two are sufficient.
    def stateHalfStartsState(self, fromNeurons, fromCA, toNeurons,toCA):
        connector = []
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            for toSynapse in range (0,self.numberHalfSynapses):
                fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
                toOffset = toSynapse + fromOffset
                toOffset = toOffset%(self.CA_SIZE-self.CA_INHIBS)
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,self.halfWeight,1.0)]
        halfConnList = pynn.FromListConnector(connector)
        projections = [
            pynn.Projection(fromNeurons, toNeurons, halfConnList, target='excitatory'),
        ]
        return projections[0]


    def stateStartsState(self, fromNeurons, fromCA, toNeurons,toCA):
        #fullWeight = fullWeight
        #print "full weight = ", fullWeight
        #brainScales constant
        connector = []
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,self.fullWeight,1.0)]
        startConnList = pynn.FromListConnector(connector)
        projections = [
            pynn.Projection(fromNeurons, toNeurons, startConnList, target='excitatory'),
        ]

        return projections[0]

    def stateInhibitsState(self, fromNeurons, fromCA, toNeurons,toCA,weight):
        #print "inhib weight = ", self.inhibWeight
        connector = []
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,weight,1.0)]
        inhibConnList = pynn.FromListConnector(connector)
        projections = [
            pynn.Projection(fromNeurons, toNeurons, inhibConnList, target='inhibitory'),
        ]

        return projections[0]

    def stateStopsState(self, fromNeurons, fromCA, toNeurons,toCA):
        #print "inhib weight = ", self.inhibWeight
        connector = []
        for fromOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
            for toOffset in range (0,self.CA_SIZE-self.CA_INHIBS):
                fromNeuron = fromOffset + (fromCA*self.CA_SIZE)
                toNeuron = toOffset + (toCA*self.CA_SIZE)
                connector = connector + [(fromNeuron,toNeuron,self.inhibWeight,1.0)]
        inhibConnList = pynn.FromListConnector(connector)
        projections = [
            pynn.Projection(fromNeurons, toNeurons, inhibConnList, target='inhibitory'),
        ]

        return projections[0]

    def makeCA(self,APop,CANum):
        #print "intraWeight = ", self.intraWeight

        intraConnector = []
        numNeurons = self.CA_SIZE-self.CA_INHIBS
        for fromOffset in range (0,numNeurons):
            for toOffset in range (0,numNeurons):
                if (fromOffset != toOffset):
                    fromNeuron = fromOffset+(CANum*self.CA_SIZE)
                    toNeuron = toOffset+(CANum*self.CA_SIZE)
                    intraConnector=intraConnector+[[fromNeuron,toNeuron,self.intraWeight,1.0]]
        intraConnList = pynn.FromListConnector(intraConnector)
        CAProjections1 = [pynn.Projection(APop, APop, intraConnList, 
                                      target='excitatory')]

        return CAProjections1[0]

    # BRAINSCALES FUNCTIONS
    #In brainscales, you have to explicitly set the weights.  This is
    #done here by passing the already created synapses and the weight
    #that they are to be set to.  This should be between 0 and 15, 
    #and inhibition needs to be determined at synapse creation.
    def setBrainScaleSWt(self,projections,weight):
        #print "projections", projections
        synapses = self.runtime.results().synapse_routing.synapses()
        #print "synapses", synapses
        proj_items = self.runtime.results().synapse_routing.synapses().find(projections)
        #print "set BS weight", weight
        #print "proj_items", proj_items
        for proj_item in proj_items:
            print "proj_item", proj_item
            synapse = proj_item.hardware_synapse()
            proxy = self.runtime.wafer()[synapse.toHICANNOnWafer()].synapses[synapse]
            proxy.weight = HICANN.SynapseWeight(weight)

    #Set all the hardware weights.  These need to be stored in arrays of 
    #projections by weight.  Each of the first 6 projections are the
    #standard types (they can be empty).  The last is specific, and
    #includes the weight.  We could condense the first 6 into the last.
    def setHardwareWeights(self,cAProjections,sourceProjections,
        halfProjections,fullProjections,stopProjections,sourceStopProjections,
                           otherProjections):
        for projections in cAProjections:
            self.setBrainScaleSWt(projections,self.intraWeight)
        for projections in sourceProjections:
            self.setBrainScaleSWt(projections,self.inputWeight)
        for projections in halfProjections:
            self.setBrainScaleSWt(projections,self.halfWeight)
        for projections in fullProjections:
            self.setBrainScaleSWt(projections,self.fullWeight)
        for projections in stopProjections:
            self.setBrainScaleSWt(projections,self.inhibWeight)
        for projections in sourceStopProjections:
            self.setBrainScaleSWt(projections,self.sourceInhibWeight)

        for anotherProjectionPair in otherProjections:
            otherWeight = anotherProjectionPair[1]
            anotherProjections = anotherProjectionPair[0]
            for projections in anotherProjections:
                self.setBrainScaleSWt(projections,otherWeight)

    #From the guidebook.  I'm not entirely sure what this does.
    def set_sthal_params(self, wafer, gmax, gmax_div):
        """
        synaptic strength:
        gmax: 0 - 1023, strongest: 1023
        gmax_div: 1 - 30, strongest: 1
        """

        # for all HICANNs in use
        for hicann in wafer.getAllocatedHicannCoordinates():

            fgs = wafer[hicann].floating_gates

            # set parameters influencing the synaptic strength
            for block in C.iter_all(C.FGBlockOnHICANN):
                fgs.setShared(block, HICANN.shared_parameter.V_gmax0, gmax)
                fgs.setShared(block, HICANN.shared_parameter.V_gmax1, gmax)
                fgs.setShared(block, HICANN.shared_parameter.V_gmax2, gmax)
                fgs.setShared(block, HICANN.shared_parameter.V_gmax3, gmax)

            for driver in C.iter_all(C.SynapseDriverOnHICANN):
                for row in C.iter_all(C.RowOnSynapseDriver):
                    wafer[hicann].synapses[driver][row].set_gmax_div(HICANN.GmaxDiv(gmax_div))

            # don't change values below
            for ii in xrange(fgs.getNoProgrammingPasses()):
                cfg = fgs.getFGConfig(C.Enum(ii))
                cfg.fg_biasn = 0
                cfg.fg_bias = 0
                fgs.setFGConfig(C.Enum(ii), cfg)

            for block in C.iter_all(C.FGBlockOnHICANN):
                fgs.setShared(block, HICANN.shared_parameter.V_dllres, 275)
                fgs.setShared(block, HICANN.shared_parameter.V_ccas, 800)

    #stuff at the beginning of the emulation
    def brainScalesHeader(self):
           self.marocco = PyMarocco()
           self.marocco.default_wafer = C.Wafer(int(os.environ.get("WAFER", 33)))
           #undone
           self.marocco.continue_despite_synapse_loss = True
           self.runtime = Runtime(self.marocco.default_wafer)
           pynn.setup(marocco=self.marocco, marocco_runtime=self.runtime)

    #stuff in the middle of the simulation
    def brainScalesMap(self, pop,hicannList):
        hicanns = []
        for hicannNum in hicannList:
            hicanns = hicanns + [C.HICANNOnWafer(C.Enum(hicannNum))]
        # hicann= C.HICANNOnWafer(C.Enum(297))
        #hicanns = [C.HICANNOnWafer(C.Enum(297)),C.HICANNOnWafer(C.Enum(296))]
        # the third args is the size of the neuron / default is 4
        #each hicann has 512 circuits, so with 8 you can have 64 neurons.
        self.marocco.manual_placement.on_hicann(pop, hicanns, 8)

        self.marocco.skip_mapping = False
        self.marocco.backend = PyMarocco.None

    #stuff just before the run starts
    def brainScalesFooter(self):
    #  ——— change low-level parameters before configuring hardware ———————————

    # call at least once
        self.set_sthal_params(self.runtime.wafer(), gmax=512, gmax_div=10)

    # Configure hardware
        self.marocco.skip_mapping = True
        self.marocco.backend = PyMarocco.Hardware
    # Full configuration during first step
        self.marocco.hicann_configurator = pysthal.ParallelHICANNv4Configurator()

